import { keywords, schema } from "@arktype/schema";
import { printable, throwParseError } from "@arktype/util";
import { parseEntry } from "./shared.js";
const stringAndSymbolicEntriesOf = (o) => [
    ...Object.entries(o),
    ...Object.getOwnPropertySymbols(o).map((k) => [k, o[k]])
];
export const parseObjectLiteral = (def, ctx) => {
    const required = [];
    const optional = [];
    // We only allow a spread operator to be used as the first key in an object
    // because to match JS behavior any keys before the spread are overwritten
    // by the values in the target object, so there'd be no useful purpose in having it
    // anywhere except for the beginning.
    // Discussion in ArkType Discord:
    // https://discord.com/channels/957797212103016458/1103023445035462678/1182814502471860334
    let hasSeenFirstKey = false;
    const entries = stringAndSymbolicEntriesOf(def);
    for (const entry of entries) {
        const result = parseEntry(entry);
        if (result.kind === "spread") {
            if (hasSeenFirstKey) {
                return throwParseError("Spread operator may only be used as the first key in an object");
            }
            const spreadNode = ctx.scope.parse(result.innerValue, ctx);
            if (spreadNode.kind !== "intersection" ||
                !spreadNode.extends(keywords.object)) {
                return throwParseError(writeInvalidSpreadTypeMessage(printable(result.innerValue)));
            }
            // For each key on spreadNode, add it to our object.
            // We filter out keys from the spreadNode that will be defined later on this same object
            // because the currently parsed definition will overwrite them.
            const requiredEntriesFromSpread = (spreadNode.props?.required ?? []).filter((e) => !entries.some(([k]) => k === e.key));
            const optionalEntriesFromSpread = (spreadNode.props?.optional ?? []).filter((e) => !entries.some(([k]) => k === e.key));
            required.push(...requiredEntriesFromSpread);
            optional.push(...optionalEntriesFromSpread);
            continue;
        }
        ctx.path.push(`${typeof result.innerKey === "symbol"
            ? `[${printable(result.innerKey)}]`
            : result.innerKey}`);
        const valueNode = ctx.scope.parse(result.innerValue, ctx);
        if (result.kind === "optional") {
            optional.push({
                key: result.innerKey,
                value: valueNode
            });
        }
        else {
            required.push({
                key: result.innerKey,
                value: valueNode
            });
        }
        ctx.path.pop();
        hasSeenFirstKey ||= true;
    }
    return schema({
        basis: "object",
        required,
        optional
    });
};
export const writeInvalidPropertyKeyMessage = (indexDef) => `Indexed key definition '${indexDef}' must be a string, number or symbol`;
export const writeInvalidSpreadTypeMessage = (def) => `Spread operand must resolve to an object literal type (was ${def})`;
//# sourceMappingURL=objectLiteral.js.map