import { BaseType, schema } from "@arktype/schema";
import { printable, throwParseError, tryParseNumber, tryParseWellFormedBigint } from "@arktype/util";
import { hasArkKind } from "../../../../util.js";
import { parseGenericArgs, writeInvalidGenericArgsMessage } from "./genericArgs.js";
export const parseUnenclosed = (s) => {
    const token = s.scanner.shiftUntilNextTerminator();
    if (token === "keyof") {
        s.addPrefix("keyof");
    }
    else {
        s.root = unenclosedToNode(s, token);
    }
};
export const parseGenericInstantiation = (name, g, s) => {
    s.scanner.shiftUntilNonWhitespace();
    const lookahead = s.scanner.shift();
    if (lookahead !== "<") {
        return s.error(writeInvalidGenericArgsMessage(name, g.parameters, []));
    }
    const parsedArgs = parseGenericArgs(name, g.parameters, s);
    const remainingChars = parsedArgs.unscanned.length;
    // set the scanner position to where the args scanner left off
    s.scanner.jumpToIndex(remainingChars === 0 ? s.scanner.length : -remainingChars);
    return g(...parsedArgs.result).root;
};
const unenclosedToNode = (s, token) => maybeParseReference(s, token) ??
    maybeParseUnenclosedLiteral(s, token) ??
    s.error(token === ""
        ? writeMissingOperandMessage(s)
        : writeUnresolvableMessage(token));
const maybeParseReference = (s, token) => {
    if (s.ctx.args?.[token]) {
        return s.ctx.args[token];
    }
    const resolution = s.ctx.scope.maybeResolve(token);
    if (resolution instanceof BaseType) {
        return resolution;
    }
    if (resolution === undefined) {
        return;
    }
    if (hasArkKind(resolution, "generic")) {
        return parseGenericInstantiation(token, resolution, s);
    }
    return throwParseError(`Unexpected resolution ${printable(resolution)}`);
};
const maybeParseUnenclosedLiteral = (s, token) => {
    const maybeNumber = tryParseNumber(token, { strict: true });
    if (maybeNumber !== undefined) {
        return schema({ unit: maybeNumber });
    }
    const maybeBigint = tryParseWellFormedBigint(token);
    if (maybeBigint !== undefined) {
        return schema({ unit: maybeBigint });
    }
};
export const writeNonSubmoduleDotMessage = (name) => `'${name}' must reference a module to be accessed using dot syntax`;
export const writeMissingSubmoduleAccessMessage = (name) => `Reference to submodule '${name}' must specify an alias`;
export const writeUnresolvableMessage = (token) => `'${token}' is unresolvable`;
export const writeMissingOperandMessage = (s) => {
    const operator = s.previousOperator();
    return operator
        ? writeMissingRightOperandMessage(operator, s.scanner.unscanned)
        : writeExpressionExpectedMessage(s.scanner.unscanned);
};
export const writeMissingRightOperandMessage = (token, unscanned = "") => `Token '${token}' requires a right operand${unscanned ? ` before '${unscanned}'` : ""}`;
export const writeExpressionExpectedMessage = (unscanned) => `Expected an expression${unscanned ? ` before '${unscanned}'` : ""}`;
//# sourceMappingURL=unenclosed.js.map