import { keywords } from "@arktype/schema";
import { isKeyOf, throwParseError, tryParseNumber } from "@arktype/util";
import { maxComparators, writeUnpairableComparatorMessage } from "../../reduce/shared.js";
import { extractDateLiteralSource, isDateLiteral } from "../operand/date.js";
export const parseBound = (s, start) => {
    const comparator = shiftComparator(s, start);
    if (s.root.kind === "unit") {
        if (typeof s.root.unit === "number") {
            s.unsetRoot();
            return s.reduceLeftBound(s.root.unit, comparator);
        }
        if (s.root.unit instanceof Date) {
            s.unsetRoot();
            const literal = `d'${s.root.description ?? s.root.unit.toISOString()}'`;
            return s.reduceLeftBound(literal, comparator);
        }
    }
    return parseRightBound(s, comparator);
};
const oneCharComparators = {
    "<": true,
    ">": true
};
export const comparatorStartChars = {
    "<": 1,
    ">": 1,
    "=": 1
};
const shiftComparator = (s, start) => s.scanner.lookaheadIs("=")
    ? `${start}${s.scanner.shift()}`
    : isKeyOf(start, oneCharComparators)
        ? start
        : s.error(singleEqualsMessage);
export const writeUnboundableMessage = (root) => `Bounded expression ${root} must be a number, string, Array, or Date`;
export const writeIncompatibleRangeMessage = (l, r) => `Bound kinds ${l} and ${r} are incompatible`;
export const writeLimitMismatchMessage = (root, limitValue) => `Limit '${limitValue}' cannot bound ${root}`;
export const getBoundKinds = (comparator, limit, root) => {
    if (root.extends(keywords.number)) {
        if (typeof limit !== "number") {
            return throwParseError(writeLimitMismatchMessage(root.toString(), limit));
        }
        return comparator === "=="
            ? ["min", "max"]
            : comparator[0] === ">"
                ? ["min"]
                : ["max"];
    }
    if (root.extends(keywords.string) || root.extends(keywords.Array)) {
        if (typeof limit !== "number") {
            return throwParseError(writeLimitMismatchMessage(root.toString(), limit));
        }
        return comparator === "=="
            ? ["minLength", "maxLength"]
            : comparator[0] === ">"
                ? ["minLength"]
                : ["maxLength"];
    }
    if (root.extends(keywords.Date)) {
        // allow either numeric or date limits
        return comparator === "=="
            ? ["after", "before"]
            : comparator[0] === ">"
                ? ["after"]
                : ["before"];
    }
    return throwParseError(writeUnboundableMessage(root.toString()));
};
export const singleEqualsMessage = `= is not a valid comparator. Use == to check for equality`;
const openLeftBoundToSchema = (leftBound) => ({
    limit: isDateLiteral(leftBound.limit)
        ? extractDateLiteralSource(leftBound.limit)
        : leftBound.limit,
    exclusive: leftBound.comparator.length === 1
});
export const parseRightBound = (s, comparator) => {
    // store the node that will be bounded
    const previousRoot = s.unsetRoot();
    const previousScannerIndex = s.scanner.location;
    s.parseOperand();
    // after parsing the next operand, use the locations to get the
    // token from which it was parsed
    const limitToken = s.scanner.sliceChars(previousScannerIndex, s.scanner.location);
    s.setRoot(previousRoot);
    const limit = tryParseNumber(limitToken) ??
        (isDateLiteral(limitToken)
            ? extractDateLiteralSource(limitToken)
            : s.error(writeInvalidLimitMessage(comparator, limitToken, "right")));
    // apply the newly-parsed right bound
    const exclusive = comparator.length === 1;
    // if the comparator is ==, both the min and max of that pair will be applied
    for (const kind of getBoundKinds(comparator, limit, previousRoot)) {
        s.constrainRoot(kind, { limit, exclusive });
    }
    if (!s.branches.leftBound) {
        return;
    }
    // if there's an open left bound, perform additional validation and apply it
    if (!isKeyOf(comparator, maxComparators)) {
        return s.error(writeUnpairableComparatorMessage(comparator));
    }
    const lowerBoundKind = getBoundKinds(s.branches.leftBound.comparator, s.branches.leftBound.limit, previousRoot);
    s.constrainRoot(lowerBoundKind[0], openLeftBoundToSchema(s.branches.leftBound));
    delete s.branches.leftBound;
};
export const writeInvalidLimitMessage = (comparator, limit, boundKind) => `Comparator ${comparator} must be ${boundKind === "left" ? "preceded" : "followed"} by a corresponding literal (was '${limit}')`;
//# sourceMappingURL=bounds.js.map