import { keywords, schema } from "@arktype/schema";
import { isArray, objectKindOrDomainOf, printable, throwParseError } from "@arktype/util";
import { writeUnsatisfiableExpressionError } from "./semantic/validate.js";
import { configureShallowDescendants, parseEntry } from "./shared.js";
import { writeMissingRightOperandMessage } from "./string/shift/operand/unenclosed.js";
export const parseTuple = (def, ctx) => maybeParseTupleExpression(def, ctx) ?? parseTupleLiteral(def, ctx);
export const parseTupleLiteral = (def, ctx) => {
    const props = [];
    let variadicIndex;
    for (let i = 0; i < def.length; i++) {
        let elementDef = def[i];
        ctx.path.push(`${i}`);
        if (typeof elementDef === "string" && elementDef.startsWith("...")) {
            elementDef = elementDef.slice(3);
            if (variadicIndex !== undefined) {
                return throwParseError(multipleVariadicMesage);
            }
            variadicIndex = i;
        }
        else if (isArray(elementDef) &&
            elementDef.length === 2 &&
            elementDef[0] === "...") {
            elementDef = elementDef[1];
            if (variadicIndex !== undefined) {
                return throwParseError(multipleVariadicMesage);
            }
            variadicIndex = i;
        }
        const parsedEntry = parseEntry([`${i}`, elementDef]);
        const value = ctx.scope.parse(parsedEntry.innerValue, ctx);
        if (variadicIndex === i) {
            if (!value.extends(keywords.Array)) {
                return throwParseError(writeNonArrayRestMessage(elementDef));
            }
            // TODO: Fix builtins.arrayIndexTypeNode()
            const elementType = value.getPath();
            // TODO: first variadic i
            props.push({ key: keywords.number, value: elementType });
        }
        else {
            props.push({
                key: {
                    name: `${i}`,
                    prerequisite: false,
                    optional: parsedEntry.kind === "optional"
                },
                value
            });
        }
        ctx.path.pop();
    }
    if (variadicIndex === undefined) {
        props.push({
            key: {
                name: "length",
                prerequisite: true,
                optional: false
            },
            value: schema({ unit: def.length })
        });
    }
    return schema(Array);
};
export const maybeParseTupleExpression = (def, ctx) => {
    const tupleExpressionResult = isIndexOneExpression(def)
        ? indexOneParsers[def[1]](def, ctx)
        : isIndexZeroExpression(def)
            ? prefixParsers[def[0]](def, ctx)
            : undefined;
    if (tupleExpressionResult) {
        return tupleExpressionResult.isNever()
            ? throwParseError(writeUnsatisfiableExpressionError(def
                .map((def) => (typeof def === "string" ? def : printable(def)))
                .join(" ")))
            : tupleExpressionResult;
    }
};
export const writeNonArrayRestMessage = (operand) => `Rest element ${typeof operand === "string" ? `'${operand}'` : ""} must be an array`;
export const multipleVariadicMesage = `A tuple may have at most one variadic element`;
export const parseKeyOfTuple = (def, ctx) => ctx.scope.parse(def[1], ctx).keyof();
const parseBranchTuple = (def, ctx) => {
    if (def[2] === undefined) {
        return throwParseError(writeMissingRightOperandMessage(def[1], ""));
    }
    const l = ctx.scope.parse(def[0], ctx);
    const r = ctx.scope.parse(def[2], ctx);
    return def[1] === "&" ? l.and(r) : l.or(r);
};
const parseArrayTuple = (def, ctx) => ctx.scope.parse(def[0], ctx).array();
const isIndexOneExpression = (def) => indexOneParsers[def[1]] !== undefined;
export const parseMorphTuple = (def, ctx) => {
    if (typeof def[2] !== "function") {
        return throwParseError(writeMalformedFunctionalExpressionMessage("=>", def[2]));
    }
    // TODO: nested morphs?
    return schema({
        in: ctx.scope.parse(def[0], ctx),
        morph: def[2]
    });
};
export const writeMalformedFunctionalExpressionMessage = (operator, value) => `${operator === ":" ? "Narrow" : "Morph"} expression requires a function following '${operator}' (was ${typeof value})`;
export const parseNarrowTuple = (def, ctx) => {
    if (typeof def[2] !== "function") {
        return throwParseError(writeMalformedFunctionalExpressionMessage(":", def[2]));
    }
    return ctx.scope
        .parse(def[0], ctx)
        .constrain("predicate", def[2]);
};
const parseAttributeTuple = (def, ctx) => configureShallowDescendants(ctx.scope.parse(def[0], ctx), def[2]);
const indexOneParsers = {
    "|": parseBranchTuple,
    "&": parseBranchTuple,
    "[]": parseArrayTuple,
    ":": parseNarrowTuple,
    "=>": parseMorphTuple,
    "@": parseAttributeTuple
};
const prefixParsers = {
    keyof: parseKeyOfTuple,
    instanceof: (def) => {
        if (typeof def[1] !== "function") {
            return throwParseError(writeInvalidConstructorMessage(objectKindOrDomainOf(def[1])));
        }
        const branches = def
            .slice(1)
            .map((ctor) => typeof ctor === "function"
            ? { proto: ctor }
            : throwParseError(writeInvalidConstructorMessage(objectKindOrDomainOf(ctor))));
        return schema(...branches);
    },
    "===": (def) => schema.units(...def.slice(1))
};
const isIndexZeroExpression = (def) => prefixParsers[def[0]] !== undefined;
export const writeInvalidConstructorMessage = (actual) => `Expected a constructor following 'instanceof' operator (was ${actual})`;
//# sourceMappingURL=tuple.js.map