import { BaseType, keywords } from "@arktype/schema";
import { domainOf, hasDomain, isThunk, map, throwParseError } from "@arktype/util";
import { createMatchParser } from "./match.js";
import { parseObject, writeBadDefinitionTypeMessage } from "./parser/definition.js";
import { parseGenericParams } from "./parser/generic.js";
import { DynamicState } from "./parser/string/reduce/dynamic.js";
import { writeMissingSubmoduleAccessMessage, writeNonSubmoduleDotMessage, writeUnresolvableMessage } from "./parser/string/shift/operand/unenclosed.js";
import { fullStringParse } from "./parser/string/string.js";
import { Type, createTypeParser, generic, validateUninstantiatedGeneric } from "./type.js";
import { addArkKind, hasArkKind } from "./util.js";
export class Scope {
    config;
    parseCache = {};
    resolutions;
    /** The set of names defined at the root-level of the scope mapped to their
     * corresponding definitions.**/
    aliases = {};
    exportedNames = [];
    ambient;
    constructor(def, config) {
        for (const k in def) {
            const parsedKey = parseScopeKey(k);
            this.aliases[parsedKey.name] = parsedKey.params.length
                ? generic(parsedKey.params, def[k], this)
                : def[k];
            if (!parsedKey.isLocal) {
                this.exportedNames.push(parsedKey.name);
            }
        }
        this.ambient = config.ambient ?? null;
        if (this.ambient) {
            // ensure exportedResolutions is populated
            this.ambient.export();
            this.resolutions = { ...this.ambient.exportedResolutions };
        }
        else {
            this.resolutions = {};
        }
        this.config = config;
    }
    static root = (aliases) => {
        return new Scope(aliases, {});
    };
    type = createTypeParser(this);
    match = createMatchParser(this);
    // TODO: decide if this API will be used for non-validated types
    declare = () => ({ type: this.type });
    scope = ((def, config = {}) => {
        return new Scope(def, {
            ambient: this.ambient,
            ...this.config,
            ...config
        });
    });
    define = (def) => def;
    toAmbient() {
        // TODO: private?
        return new Scope(this.aliases, {
            ambient: this,
            ...this.config
        });
    }
    // TODO: name?
    get(name) {
        return this.export()[name];
    }
    createRootContext(input) {
        return {
            path: [],
            scope: this,
            ...input
        };
    }
    parseDefinition(def, input) {
        return this.parse(def, this.createRootContext(input));
    }
    parse(def, ctx) {
        if (typeof def === "string") {
            if (ctx.args !== undefined) {
                // we can only rely on the cache if there are no contextual
                // resolutions like "this" or generic args
                return this.parseString(def, ctx);
            }
            if (!this.parseCache[def]) {
                this.parseCache[def] = this.parseString(def, ctx);
            }
            return this.parseCache[def];
        }
        return hasDomain(def, "object")
            ? parseObject(def, ctx)
            : throwParseError(writeBadDefinitionTypeMessage(domainOf(def)));
    }
    parseString(def, ctx) {
        return (this.maybeResolveNode(def) ??
            ((def.endsWith("[]") &&
                this.maybeResolveNode(def.slice(0, -2))?.array()) ||
                fullStringParse(new DynamicState(def, ctx))));
    }
    maybeResolve(name) {
        const cached = this.resolutions[name];
        if (cached) {
            return cached;
        }
        let def = this.aliases[name];
        if (!def) {
            return this.maybeResolveSubalias(name);
        }
        if (isThunk(def) && !hasArkKind(def, "generic")) {
            def = def();
        }
        // TODO: initialize here?
        const resolution = hasArkKind(def, "generic")
            ? validateUninstantiatedGeneric(def)
            : hasArkKind(def, "module")
                ? throwParseError(writeMissingSubmoduleAccessMessage(name))
                : this.parseDefinition(def, this.createRootContext({ baseName: name, args: {} }));
        this.resolutions[name] = resolution;
        return resolution;
    }
    /** If name is a valid reference to a submodule alias, return its resolution  */
    maybeResolveSubalias(name) {
        const dotIndex = name.indexOf(".");
        if (dotIndex === -1) {
            return;
        }
        const dotPrefix = name.slice(0, dotIndex);
        const prefixDef = this.aliases[dotPrefix];
        if (hasArkKind(prefixDef, "module")) {
            const resolution = prefixDef[name.slice(dotIndex + 1)]?.root;
            // if the first part of name is a submodule but the suffix is
            // unresolvable, we can throw immediately
            if (!resolution) {
                return throwParseError(writeUnresolvableMessage(name));
            }
            this.resolutions[name] = resolution;
            return resolution;
        }
        if (prefixDef !== undefined) {
            return throwParseError(writeNonSubmoduleDotMessage(dotPrefix));
        }
        // if the name includes ".", but the prefix is not an alias, it
        // might be something like a decimal literal, so just fall through to return
    }
    maybeResolveNode(name) {
        const result = this.maybeResolve(name);
        return result instanceof BaseType ? result : undefined;
    }
    import(...names) {
        return addArkKind(map(this.export(...names), (alias, value) => [
            `#${alias}`,
            value
        ]), "module");
    }
    bindThis() {
        // TODO: fix
        return { this: keywords.unknown };
    }
    exportedResolutions;
    exportCache;
    export(...names) {
        if (!this.exportCache) {
            this.exportCache = {};
            for (const name of this.exportedNames) {
                let def = this.aliases[name];
                if (hasArkKind(def, "generic")) {
                    this.exportCache[name] = def;
                    continue;
                }
                // TODO: thunk generics?
                // handle generics before invoking thunks, since they use
                // varargs they will incorrectly be considered thunks
                if (isThunk(def)) {
                    def = def();
                }
                if (hasArkKind(def, "module")) {
                    this.exportCache[name] = def;
                }
                else {
                    this.exportCache[name] = new Type(this.maybeResolve(name), this);
                }
            }
            this.exportedResolutions = resolutionsOfModule(this.exportCache);
            Object.assign(this.resolutions, this.exportedResolutions);
        }
        const namesToExport = names.length ? names : this.exportedNames;
        return addArkKind(map(namesToExport, (_, name) => [name, this.exportCache[name]]), "module");
    }
}
const resolutionsOfModule = (typeSet) => {
    const result = {};
    for (const k in typeSet) {
        const v = typeSet[k];
        if (hasArkKind(v, "module")) {
            const innerResolutions = resolutionsOfModule(v);
            const prefixedResolutions = map(innerResolutions, (innerK, innerV) => [
                `${k}.${innerK}`,
                innerV
            ]);
            Object.assign(result, prefixedResolutions);
        }
        else if (hasArkKind(v, "generic")) {
            result[k] = v;
        }
        else {
            result[k] = v.root;
        }
    }
    return result;
};
export const writeShallowCycleErrorMessage = (name, seen) => `Alias '${name}' has a shallow resolution cycle: ${[...seen, name].join(":")}`;
export const writeDuplicateNameMessage = (name) => `Duplicate name '${name}'`;
export const parseScopeKey = (k) => {
    const isLocal = k[0] === "#";
    const name = isLocal ? k.slice(1) : k;
    const firstParamIndex = k.indexOf("<");
    if (firstParamIndex === -1) {
        return {
            isLocal,
            name,
            params: []
        };
    }
    if (k.at(-1) !== ">") {
        throwParseError(`'>' must be the last character of a generic declaration in a scope`);
    }
    return {
        isLocal,
        name: name.slice(0, firstParamIndex),
        params: parseGenericParams(k.slice(firstParamIndex + 1, -1))
    };
};
//# sourceMappingURL=scope.js.map