import { inferred, keywords } from "@arktype/schema";
import { Callable, map } from "@arktype/util";
import { parseGenericParams } from "./parser/generic.js";
import { configureShallowDescendants } from "./parser/shared.js";
import { arkKind } from "./util.js";
export const createTypeParser = (scope) => {
    const parser = (...args) => {
        if (args.length === 1) {
            // treat as a simple definition
            return new Type(args[0], scope);
        }
        if (args.length === 2 &&
            typeof args[0] === "string" &&
            args[0][0] === "<" &&
            args[0].at(-1) === ">") {
            // if there are exactly two args, the first of which looks like <${string}>,
            // treat as a generic
            const params = parseGenericParams(args[0].slice(1, -1));
            const def = args[1];
            return validateUninstantiatedGeneric(generic(params, def, scope));
        }
        // otherwise, treat as a tuple expression. technically, this also allows
        // non-expression tuple definitions to be parsed, but it's not a supported
        // part of the API as specified by the associated types
        return new Type(args, scope);
    };
    return parser;
};
export class Type extends Callable {
    definition;
    scope;
    root;
    allows;
    description;
    json;
    constructor(definition, scope) {
        const root = parseTypeRoot(definition, scope);
        super(root.apply, root);
        this.definition = definition;
        this.scope = scope;
        this.root = root;
        this.allows = root.allows;
        this.json = root.json;
        this.description = this.root.description;
    }
    configure(configOrDescription) {
        return new Type(configureShallowDescendants(this.root, configOrDescription), this.scope);
    }
    describe(description) {
        return this.configure(description);
    }
    // TODO: should return out
    from(literal) {
        return literal;
    }
    // TODO: Morph intersections, ordering
    and(def) {
        return new Type(this.root.and(parseTypeRoot(def, this.scope)), this.scope);
    }
    or(def) {
        return new Type(this.root.or(parseTypeRoot(def, this.scope)), this.scope);
    }
    morph(morph, outValidator) {
        // TODO: tuple expression for out validator
        outValidator;
        return this;
        // return new Type(
        //     this.root.constrain("morph", morph),
        //     this.scope
        // ) as never
    }
    // TODO: based on below, should maybe narrow morph output if used after
    narrow(def) {
        return new Type(this.root.constrain("predicate", def), this.scope);
    }
    array() {
        return new Type(this.root.array(), this.scope);
    }
    keyof() {
        return new Type(this.root.keyof(), this.scope);
    }
    assert(data) {
        const result = this.call(null, data);
        return result.errors ? result.errors.throw() : result.out;
    }
    equals(other) {
        return this.root.equals(parseTypeRoot(other, this.scope));
    }
    extends(other) {
        return this.root.extends(parseTypeRoot(other, this.scope));
    }
    inCache;
    get in() {
        this.inCache ??= new Type(this.root.in, this.scope);
        return this.inCache;
    }
    outCache;
    get out() {
        this.outCache ??= new Type(this.root.out, this.scope);
        return this.outCache;
    }
    toString() {
        return this.description;
    }
}
const parseTypeRoot = (def, scope, args) => scope.parseDefinition(def, {
    args: args ?? scope.bindThis(),
    baseName: "type"
});
export const validateUninstantiatedGeneric = (g) => {
    // the unconstrained instantiation of the generic is not used for now
    // other than to eagerly validate that the def does not contain any errors
    g.scope.parseDefinition(g.definition, 
    // once we support constraints on generic parameters, we'd use
    // the base type here: https://github.com/arktypeio/arktype/issues/796
    {
        baseName: "generic",
        args: map(g.parameters, (_, name) => [name, keywords.unknown])
    });
    return g;
};
export const generic = (parameters, definition, scope) => {
    return Object.assign((...args) => {
        const argNodes = map(parameters, (i, param) => [
            param,
            parseTypeRoot(args[i], scope)
        ]);
        const root = parseTypeRoot(definition, scope, argNodes);
        return new Type(root, scope);
    }, {
        [arkKind]: "generic",
        parameters,
        definition,
        scope
        // $ is only needed at compile-time
    });
};
//# sourceMappingURL=type.js.map